"""Compare module."""
import warnings
from itertools import combinations
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from scipy import stats
from statsmodels.stats.multitest import multipletests
from tqdm import tqdm

from abby.diagnose import sample_ratio_mismatch
from abby.utils import Ratio


def compare_multiple(
    data: pd.DataFrame,
    variants: List[str],
    metrics: List[Union[str, Ratio]],
    method: Optional[str] = "bonferroni",
):
    assert "variant_name" in data.columns, "Rename the variant column to `variant_name`"
    assert set(variants) == set(
        data["variant_name"].unique()
    ), "One or more variants are not found in the data"
    if sample_ratio_mismatch(data["variant_name"]) < 0.05:
        warnings.warn("There are sample ratio mismatch, your variants are not balance")

    results = []

    for metric in metrics:
        if isinstance(metric, Ratio):
            print("ratio")
            result_df = compare_delta(
                data, variants, metric.numerator, metric.denominator, method=method
            )
            result_df["metric"] = metric.name
            results.append(result_df)

        elif isinstance(metric, str):
            result_df = compare_ttest(data, variants, metric, method=method)
            result_df["metric"] = metric
            results.append(result_df)
        else:
            raise ValueError(f"Unknown type: {type(metric)}")

    return pd.concat(results).set_index("metric")


def compare_ttest(
    data: pd.DataFrame,
    variants: List[str],
    numerator: str,
    method: Optional[str] = "bonferroni",
):
    assert "variant_name" in data.columns, "Rename the variant column to `variant_name`"
    assert set(variants) == set(
        data["variant_name"].unique()
    ), "One or more variants are not found in the data"
    if sample_ratio_mismatch(data["variant_name"]) < 0.05:
        warnings.warn("There are sample ratio mismatch, your variants are not balance")

    if len(variants) == 2:
        a_name, b_name = variants

        a_num = data.loc[data["variant_name"] == a_name, numerator]
        b_num = data.loc[data["variant_name"] == b_name, numerator]

        return _compare_ttest(a_num, a_name, b_num, b_name)

    elif len(variants) > 2:
        pairs = combinations(variants, 2)
        results = []
        for (a_name, b_name) in pairs:
            a_num = data.loc[data["variant_name"] == a_name, numerator]
            b_num = data.loc[data["variant_name"] == b_name, numerator]
            results.append(_compare_ttest(a_num, a_name, b_num, b_name))
        results_df = pd.concat(results)
        results_df["p_values_corrected"] = multipletests(
            results_df["p_values"].values, method=method
        )[1]
        return results_df


def _compare_ttest(
    a: np.array, a_name: str, b: np.array, b_name: str
) -> Dict[str, float]:
    a_size, b_size = len(a), len(b)
    a_mean, b_mean = np.mean(a), np.mean(b)

    a_var, b_var = np.var(a, ddof=1), np.var(b, ddof=1)

    delta = b_mean - a_mean
    _, p_values = stats.ttest_ind(a, b, equal_var=False)
    stde = 1.96 * np.sqrt(a_var / a_size + b_var / b_size)

    result = dict(
        A=a_name,
        B=b_name,
        mean_A=a_mean,
        mean_B=b_mean,
        var_A=a_var,
        var_B=b_var,
        absolute_difference=delta,
        lower_bound=delta - stde,
        upper_bound=delta + stde,
        p_values=p_values,
    )
    return pd.DataFrame([result])


def compare_bootstrap_delta(
    data: pd.DataFrame,
    variants: List[str],
    numerator: str,
    denominator: Optional[str] = "",
    **kwargs,
):
    assert "variant_name" in data.columns, "Rename the variant column to `variant_name`"
    if sample_ratio_mismatch(data["variant_name"]) < 0.05:
        warnings.warn("There are sample ratio mismatch, your variants are not balance")

    ctrl, exp = variants

    control_num = data.loc[data["variant_name"] == ctrl, numerator].values
    control_denom = data.loc[data["variant_name"] == ctrl, denominator].values
    exp_num = data.loc[data["variant_name"] == exp, numerator].values
    exp_denom = data.loc[data["variant_name"] == exp, denominator].values

    return _compare_bootstrap_delta(
        control_num,
        control_denom,
        exp_num,
        exp_denom,
        **kwargs,
    )


def compare_delta(
    data: pd.DataFrame,
    variants: List[str],
    numerator: str,
    denominator: str,
    method: Optional[str] = "bonferroni",
) -> Dict[str, float]:
    assert "variant_name" in data.columns, "Rename the variant column to `variant_name`"
    assert set(variants) == set(
        data["variant_name"].unique()
    ), "One or more variants are not found in the data"
    if sample_ratio_mismatch(data["variant_name"]) < 0.05:
        warnings.warn("There are sample ratio mismatch, your variants are not balance")

    if len(variants) == 2:
        a_name, b_name = variants

        a_num = data.loc[data["variant_name"] == a_name, numerator]
        a_denom = data.loc[data["variant_name"] == a_name, denominator]
        b_num = data.loc[data["variant_name"] == b_name, numerator]
        b_denom = data.loc[data["variant_name"] == b_name, denominator]

        return _compare_delta(a_num, a_denom, a_name, b_num, b_denom, b_name)

    elif len(variants) > 2:
        pairs = combinations(variants, 2)
        results = []
        for (a_name, b_name) in pairs:
            a_num = data.loc[data["variant_name"] == a_name, numerator]
            a_denom = data.loc[data["variant_name"] == a_name, denominator]
            b_num = data.loc[data["variant_name"] == b_name, numerator]
            b_denom = data.loc[data["variant_name"] == b_name, denominator]
            results.append(
                _compare_delta(
                    a_num,
                    a_denom,
                    a_name,
                    b_num,
                    b_denom,
                    b_name,
                )
            )
        results_df = pd.concat(results)
        results_df["p_values_corrected"] = multipletests(
            results_df["p_values"].values, method=method
        )[1]
        return results_df


def _compare_bootstrap_delta(
    control_num: np.array,
    control_denom: np.array,
    exp_num: np.array,
    exp_denom: np.array,
    n_bootstrap: int = 10_000,
):
    n_users_a, n_users_b = len(control_num), len(exp_num)
    n_users = n_users_a + n_users_b
    bs_observed = []

    for _ in tqdm(range(n_bootstrap)):
        conversion = np.hstack((control_num, exp_num))
        session = np.hstack((control_denom, exp_denom))

        assignments = np.random.choice(n_users, n_users, replace=True)
        ctrl_idxs = assignments[: int(n_users / 2)]
        test_idxs = assignments[int(n_users / 2) :]

        bs_control_denom = session[ctrl_idxs]
        bs_denominator_exp = session[test_idxs]
        bs_control_num = conversion[ctrl_idxs]
        bs_exp_num = conversion[test_idxs]

        bs_observed.append(
            bs_exp_num.sum() / bs_denominator_exp.sum()
            - bs_control_num.sum() / bs_control_denom.sum()
        )

    observed_diffs = (
        exp_num.sum() / exp_denom.sum() - control_num.sum() / control_denom.sum()
    )

    lower_bound, upper_bound = _confidence_interval_bootstrap(
        control_num,
        control_denom,
        exp_num,
        exp_denom,
        n_bootstrap,
    )
    p_values = 2 * (1 - (np.abs(observed_diffs) > np.array(bs_observed)).mean())
    return dict(
        control_mean=control_num.sum() / control_denom.sum(),
        experiment_mean=exp_num.sum() / exp_denom.sum(),
        control_var=ratio_variance(control_num, control_denom),
        experiment_var=ratio_variance(exp_num, exp_denom),
        absolute_difference=observed_diffs,
        lower_bound=lower_bound,
        upper_bound=upper_bound,
        p_values=p_values,
    )


def _confidence_interval_bootstrap(
    numerator_ctrl: np.array,
    denominator_ctrl: np.array,
    numerator_exp: np.array,
    denominator_exp: np.array,
    n_bootstrap: int,
):
    bs_observed = []

    for _ in tqdm(range(n_bootstrap)):
        ctrl_idxs = np.random.choice(
            len(numerator_ctrl), len(numerator_ctrl), replace=True
        )
        exp_idxs = np.random.choice(
            len(numerator_exp), len(numerator_exp), replace=True
        )

        bs_denominator_ctrl = denominator_ctrl[ctrl_idxs]
        bs_denominator_exp = denominator_exp[exp_idxs]
        bs_numerator_ctrl = numerator_ctrl[ctrl_idxs]
        bs_numerator_exp = numerator_exp[exp_idxs]

        bs_observed.append(
            bs_numerator_exp.sum() / bs_denominator_exp.sum()
            - bs_numerator_ctrl.sum() / bs_denominator_ctrl.sum()
        )
    return np.percentile(bs_observed, [2.5, 97.5])


def _compare_delta(
    a_num: np.array,
    a_denom: np.array,
    a_name: str,
    b_num: np.array,
    b_denom: np.array,
    b_name: str,
) -> Dict[str, float]:

    a_size = len(a_num)
    b_size = len(b_num)

    a_var = ratio_variance(a_num, a_denom)
    b_var = ratio_variance(b_num, b_denom)

    a_mean = a_num.sum() / a_denom.sum()
    b_mean = b_num.sum() / b_denom.sum()

    delta = b_mean - a_mean
    stde = 1.96 * np.sqrt(a_var / a_size + b_var / b_size)

    z_scores = np.abs(delta) / np.sqrt(a_var / a_size + b_var / b_size)
    p_values = stats.norm.sf(abs(z_scores)) * 2

    result = dict(
        A=a_name,
        B=b_name,
        mean_A=a_mean,
        mean_B=b_mean,
        var_A=a_var,
        var_B=b_var,
        absolute_difference=delta,
        lower_bound=delta - stde,
        upper_bound=delta + stde,
        p_values=p_values,
    )
    return pd.DataFrame([result])


def ratio_variance(num: np.array, denom: np.array) -> float:
    """
    Reference:
    https://www.stat.cmu.edu/~hseltman/files/ratio.pdf
    """
    assert len(num) == len(
        denom
    ), f"Different length between num: {len(num)} and denom: {len(denom)}"
    denom_mean = np.mean(denom)
    num_mean = np.mean(num)
    denom_variance = np.var(denom, ddof=1)
    num_variance = np.var(num, ddof=1)
    denom_num_covariance = np.cov(denom, num, ddof=1)[0][1]
    return (
        (num_variance) / (denom_mean**2)
        - 2 * num_mean * denom_num_covariance / (denom_mean**3)
        + (num_mean**2) * (denom_variance) / (denom_mean**4)
    )



import pandas as pd
from scipy import stats


def sample_ratio_mismatch(series: pd.Series) -> float:
    observed = series.value_counts().values
    n = len(observed)
    expected = [len(series) / n] * n
    pvalue = stats.chisquare(observed, expected).pvalue
    return pvalue


def check_null():
    raise NotImplementedError



"""Utils module."""


class Ratio:
    numerator: str
    denominator: str

    def __init__(self, numerator, denominator):
        self.numerator = numerator
        self.denominator = denominator

    @property
    def name(self):
        return f"{self.numerator}/{self.denominator}"

    def __repr__(self):
        return f"{self.numerator}/{self.denominator}"



import pandas as pd


def load_dataset(name: str, **kwargs) -> pd.DataFrame:
    url = f"https://raw.githubusercontent.com/farhanreynaldo/abby/main/abby/tests/datasets/{name}.csv"  # noqa

    df = pd.read_csv(url, **kwargs)
    return df



from typing import Callable

import numpy as np
import pandas as pd
import pytest

from abby.compare import compare_ttest


@pytest.fixture
def sleeps() -> pd.DataFrame:
    data = pd.read_csv("abby/tests/datasets/sleeps.csv").assign(
        variant_name=lambda df: df["group"].map({1: "a", 2: "b"})
    )
    return data


@pytest.fixture
def sleeps_wrong(sleeps):
    return sleeps.rename(columns={"variant_name": "group"})


@pytest.fixture
def sleeps_mismatch_ratio(sleeps: Callable) -> pd.DataFrame:
    return sleeps.assign(variant_name=np.repeat(["a", "b"], [18, 2]))


@pytest.fixture
def click_multi_df() -> pd.DataFrame:
    data = pd.read_csv("abby/tests/datasets/click_impression_3_variants.csv")
    return data


class TestCompareTtest:
    def test_ttest_result(self, sleeps: Callable):
        result = compare_ttest(sleeps, ["a", "b"], "extra").loc[0]
        assert result["mean_A"] == pytest.approx(0.75, rel=4)
        assert result["mean_B"] == pytest.approx(2.33, rel=4)
        assert result["var_A"] == pytest.approx(3.200556, rel=4)
        assert result["var_B"] == pytest.approx(4.009000, rel=4)
        assert result["absolute_difference"] == pytest.approx(1.5800, rel=4)
        assert result["lower_bound"] == pytest.approx(-0.2054832, rel=4)
        assert result["upper_bound"] == pytest.approx(3.3654832, rel=4)
        assert result["p_values"] == pytest.approx(0.07939, rel=4)

    def test_ttest_wrong_variant_column_name(self, sleeps_wrong: Callable):
        with pytest.raises(AssertionError):
            compare_ttest(sleeps_wrong, ["a", "b"], "extra")

    def test_ttest_raise_warning_mismatch_ratio(self, sleeps_mismatch_ratio: Callable):
        with pytest.warns(UserWarning):
            compare_ttest(sleeps_mismatch_ratio, ["a", "b"], "extra")


class TestCompareTtestMultipleVariants:
    def test_ttest_result_multiple_variants(self, click_multi_df: Callable):
        result = compare_ttest(
            click_multi_df, ["control", "experiment_A", "experiment_B"], "click"
        )
        control_and_A = result.loc[
            lambda df: (df["A"] == "control") & (df["B"] == "experiment_A")
        ].loc[0]
        control_and_B = result.loc[
            lambda df: (df["A"] == "control") & (df["B"] == "experiment_B")
        ].loc[0]
        assert control_and_A["mean_A"] == pytest.approx(0.45, rel=4)
        assert control_and_A["mean_B"] == pytest.approx(0.63, rel=4)
        assert control_and_A["var_A"] == pytest.approx(0.351010, rel=4)
        assert control_and_A["var_B"] == pytest.approx(0.740505, rel=4)
        assert control_and_A["absolute_difference"] == pytest.approx(0.18, rel=4)
        assert control_and_A["lower_bound"] == pytest.approx(-0.024772, rel=4)
        assert control_and_A["upper_bound"] == pytest.approx(0.384772, rel=4)
        assert control_and_A["p_values"] == pytest.approx(8.666844e-02, rel=4)

        assert control_and_B["mean_A"] == pytest.approx(0.45, rel=4)
        assert control_and_B["mean_B"] == pytest.approx(0.98, rel=4)
        assert control_and_B["var_A"] == pytest.approx(0.351010, rel=4)
        assert control_and_B["var_B"] == pytest.approx(0.666263, rel=4)
        assert control_and_B["absolute_difference"] == pytest.approx(0.53, rel=4)
        assert control_and_B["lower_bound"] == pytest.approx(0.332315, rel=4)
        assert control_and_B["upper_bound"] == pytest.approx(0.727685, rel=4)
        assert control_and_B["p_values"] == pytest.approx(4.145101e-07, rel=4)



from typing import Callable

import numpy as np
import pandas as pd
import pytest

from abby.compare import compare_multiple
from abby.utils import Ratio


@pytest.fixture
def click_df() -> pd.DataFrame:
    data = pd.read_csv("abby/tests/datasets/click_impression.csv")
    return data


@pytest.fixture
def click_multi_df() -> pd.DataFrame:
    data = pd.read_csv("abby/tests/datasets/click_impression_3_variants.csv")
    return data


@pytest.fixture
def click_df_wrong(click_df):
    return click_df.rename(columns={"variant_name": "group"})


@pytest.fixture
def click_df_mismatch_ratio(click_df: Callable) -> pd.DataFrame:
    return click_df.assign(variant_name=np.repeat(["control", "experiment"], [180, 20]))


class TestCompareMultiple:
    def test_multiple_result(self, click_df: Callable):
        result = compare_multiple(
            click_df, ["control", "experiment"], ["click", Ratio("click", "impression")]
        )
        assert result.loc["click"]["mean_A"] == pytest.approx(0.45, rel=4)
        assert result.loc["click"]["mean_B"] == pytest.approx(0.63, rel=4)
        assert result.loc["click"]["var_A"] == pytest.approx(0.35101, rel=4)
        assert result.loc["click"]["var_B"] == pytest.approx(0.74050, rel=4)
        assert result.loc["click"]["absolute_difference"] == pytest.approx(0.18, rel=4)
        assert result.loc["click"]["lower_bound"] == pytest.approx(-0.02618, rel=4)
        assert result.loc["click"]["upper_bound"] == pytest.approx(0.38618, rel=4)
        assert result.loc["click"]["p_values"] == pytest.approx(0.08666, rel=4)

        name = "click/impression"
        assert result.loc[name]["mean_A"] == pytest.approx(0.135135, rel=4)
        assert result.loc[name]["mean_B"] == pytest.approx(0.162371, rel=4)
        assert result.loc[name]["var_A"] == pytest.approx(0.0295697, rel=4)
        assert result.loc[name]["var_B"] == pytest.approx(0.0475355, rel=4)
        assert result.loc[name]["absolute_difference"] == pytest.approx(0.027236, rel=4)
        assert result.loc[name]["lower_bound"] == pytest.approx(-0.027189, rel=4)
        assert result.loc[name]["upper_bound"] == pytest.approx(0.081661, rel=4)
        assert result.loc[name]["p_values"] == pytest.approx(0.326668, rel=4)

    def test_multiple_wrong_variant_column_name(self, click_df_wrong: Callable):
        with pytest.raises(AssertionError):
            compare_multiple(
                click_df_wrong,
                ["control", "experiment"],
                ["click", Ratio("click", "impression")],
            )

    def test_multiple_raise_warning_mismatch_ratio(
        self, click_df_mismatch_ratio: Callable
    ):
        with pytest.warns(UserWarning):
            compare_multiple(
                click_df_mismatch_ratio,
                ["control", "experiment"],
                ["click", Ratio("click", "impression")],
            )


class TestCompareMultipleMultipleVariants:
    def test_multiple_result_multiple_variants(self, click_multi_df: Callable):
        result = compare_multiple(
            click_multi_df,
            ["control", "experiment_A", "experiment_B"],
            ["click", Ratio("click", "impression")],
        )
        control_and_A = result.loc[
            lambda df: (df["A"] == "control") & (df["B"] == "experiment_A")
        ]
        control_and_B = result.loc[
            lambda df: (df["A"] == "control") & (df["B"] == "experiment_B")
        ]
        assert control_and_A.loc["click"]["mean_A"] == pytest.approx(0.45, rel=4)
        assert control_and_A.loc["click"]["mean_B"] == pytest.approx(0.63, rel=4)
        assert control_and_A.loc["click"]["var_A"] == pytest.approx(0.35101, rel=4)
        assert control_and_A.loc["click"]["var_B"] == pytest.approx(0.74050, rel=4)
        assert control_and_A.loc["click"]["absolute_difference"] == pytest.approx(
            0.18, rel=4
        )
        assert control_and_A.loc["click"]["lower_bound"] == pytest.approx(
            -0.02618, rel=4
        )
        assert control_and_A.loc["click"]["upper_bound"] == pytest.approx(
            0.38618, rel=4
        )
        assert control_and_A.loc["click"]["p_values"] == pytest.approx(0.08666, rel=4)

        name = "click/impression"
        assert control_and_A.loc[name]["mean_A"] == pytest.approx(0.135135, rel=4)
        assert control_and_A.loc[name]["mean_B"] == pytest.approx(0.162371, rel=4)
        assert control_and_A.loc[name]["var_A"] == pytest.approx(0.0295697, rel=4)
        assert control_and_A.loc[name]["var_B"] == pytest.approx(0.0475355, rel=4)
        assert control_and_A.loc[name]["absolute_difference"] == pytest.approx(
            0.027236, rel=4
        )
        assert control_and_A.loc[name]["lower_bound"] == pytest.approx(-0.027189, rel=4)
        assert control_and_A.loc[name]["upper_bound"] == pytest.approx(0.081661, rel=4)
        assert control_and_A.loc[name]["p_values"] == pytest.approx(0.326668, rel=4)

        assert control_and_B.loc["click"]["mean_A"] == pytest.approx(0.45, rel=4)
        assert control_and_B.loc["click"]["mean_B"] == pytest.approx(0.98, rel=4)
        assert control_and_B.loc["click"]["var_A"] == pytest.approx(0.351010, rel=4)
        assert control_and_B.loc["click"]["var_B"] == pytest.approx(0.666263, rel=4)
        assert control_and_B.loc["click"]["absolute_difference"] == pytest.approx(
            0.53, rel=4
        )
        assert control_and_B.loc["click"]["lower_bound"] == pytest.approx(
            0.332315, rel=4
        )
        assert control_and_B.loc["click"]["upper_bound"] == pytest.approx(
            0.727685, rel=4
        )
        assert control_and_B.loc["click"]["p_values"] == pytest.approx(
            4.145101e-07, rel=4
        )
        name = "click/impression"
        assert control_and_B.loc[name]["mean_A"] == pytest.approx(0.135135, rel=4)
        assert control_and_B.loc[name]["mean_B"] == pytest.approx(0.259259, rel=4)
        assert control_and_B.loc[name]["var_A"] == pytest.approx(0.0295697, rel=4)
        assert control_and_B.loc[name]["var_B"] == pytest.approx(0.054965, rel=4)
        assert control_and_B.loc[name]["absolute_difference"] == pytest.approx(
            0.124124, rel=4
        )
        assert control_and_B.loc[name]["lower_bound"] == pytest.approx(0.067138, rel=4)
        assert control_and_B.loc[name]["upper_bound"] == pytest.approx(0.181111, rel=4)
        assert control_and_B.loc[name]["p_values"] == pytest.approx(0.00002, rel=4)



"""Seamless A/B testing with Abby."""
from abby.compare import (  # noqa
    compare_bootstrap_delta,
    compare_delta,
    compare_multiple,
    compare_ttest,
)
from abby.utils import Ratio  # noqa

__version__ = "0.1.1"



from typing import Callable

import numpy as np
import pandas as pd
import pytest

from abby.compare import compare_delta


@pytest.fixture
def click_df() -> pd.DataFrame:
    data = pd.read_csv("abby/tests/datasets/click_impression.csv")
    return data


@pytest.fixture
def click_multi_df() -> pd.DataFrame:
    data = pd.read_csv("abby/tests/datasets/click_impression_3_variants.csv")
    return data


@pytest.fixture
def click_df_wrong(click_df: Callable) -> pd.DataFrame:
    return click_df.rename(columns={"variant_name": "group"})


@pytest.fixture
def click_df_mismatch_ratio(click_df: Callable) -> pd.DataFrame:
    return click_df.assign(variant_name=np.repeat(["control", "experiment"], [180, 20]))


class TestCompareDelta:
    def test_delta_result(self, click_df: Callable):
        result = compare_delta(
            click_df, ["control", "experiment"], "click", "impression"
        ).loc[0]
        assert result["mean_A"] == pytest.approx(0.135135, rel=4)
        assert result["mean_B"] == pytest.approx(0.162371, rel=4)
        assert result["var_A"] == pytest.approx(0.0295697, rel=4)
        assert result["var_B"] == pytest.approx(0.0475355, rel=4)
        assert result["absolute_difference"] == pytest.approx(0.027236, rel=4)
        assert result["lower_bound"] == pytest.approx(-0.027189, rel=4)
        assert result["upper_bound"] == pytest.approx(0.081661, rel=4)
        assert result["p_values"] == pytest.approx(0.326668, rel=4)

    def test_delta_wrong_variant_column_name(self, click_df_wrong: Callable):
        with pytest.raises(AssertionError):
            compare_delta(
                click_df_wrong, ["control", "experiment"], "click", "impression"
            )

    def test_delta_raise_warning_mismatch_ratio(
        self, click_df_mismatch_ratio: Callable
    ):
        with pytest.warns(UserWarning):
            compare_delta(
                click_df_mismatch_ratio,
                ["control", "experiment"],
                "click",
                "impression",
            )


class TestCompareDeltaMultipleVariants:
    def test_delta_result_multiple_variants(self, click_multi_df: Callable):
        result = compare_delta(
            click_multi_df,
            ["control", "experiment_A", "experiment_B"],
            "click",
            "impression",
        )
        control_and_A = result.loc[
            lambda df: (df["A"] == "control") & (df["B"] == "experiment_A")
        ].loc[0]
        control_and_B = result.loc[
            lambda df: (df["A"] == "control") & (df["B"] == "experiment_B")
        ].loc[0]
        assert control_and_A["mean_A"] == pytest.approx(0.135135, rel=4)
        assert control_and_A["mean_B"] == pytest.approx(0.162371, rel=4)
        assert control_and_A["var_A"] == pytest.approx(0.0295697, rel=4)
        assert control_and_A["var_B"] == pytest.approx(0.0475355, rel=4)
        assert control_and_A["absolute_difference"] == pytest.approx(0.027236, rel=4)
        assert control_and_A["lower_bound"] == pytest.approx(-0.027189, rel=4)
        assert control_and_A["upper_bound"] == pytest.approx(0.081661, rel=4)
        assert control_and_A["p_values"] == pytest.approx(0.326668, rel=4)

        assert control_and_B["mean_A"] == pytest.approx(0.135135, rel=4)
        assert control_and_B["mean_B"] == pytest.approx(0.259259, rel=4)
        assert control_and_B["var_A"] == pytest.approx(0.0295697, rel=4)
        assert control_and_B["var_B"] == pytest.approx(0.054965, rel=4)
        assert control_and_B["absolute_difference"] == pytest.approx(0.124124, rel=4)
        assert control_and_B["lower_bound"] == pytest.approx(0.067138, rel=4)
        assert control_and_B["upper_bound"] == pytest.approx(0.181111, rel=4)
        assert control_and_B["p_values"] == pytest.approx(0.00002, rel=4)



from typing import Callable

import numpy as np
import pandas as pd
import pytest

from abby.compare import compare_bootstrap_delta


@pytest.fixture
def click_df() -> pd.DataFrame:
    data = pd.read_csv("abby/tests/datasets/click_impression.csv")
    return data


@pytest.fixture
def click_df_wrong(click_df):
    return click_df.rename(columns={"variant_name": "group"})


@pytest.fixture
def click_df_mismatch_ratio(click_df: Callable) -> pd.DataFrame:
    return click_df.assign(variant_name=np.repeat(["control", "experiment"], [180, 20]))


class TestCompareBootstrap:
    def test_bootstrap_result(self, click_df: Callable):
        result = compare_bootstrap_delta(
            click_df, ["control", "experiment"], "click", "impression"
        )
        assert result["control_mean"] == pytest.approx(0.135135, rel=2)
        assert result["experiment_mean"] == pytest.approx(0.162371, rel=2)
        assert result["control_var"] == pytest.approx(0.0295697, rel=2)
        assert result["experiment_var"] == pytest.approx(0.0475355, rel=2)
        assert result["absolute_difference"] == pytest.approx(0.027236, rel=2)
        assert result["lower_bound"] == pytest.approx(-0.027189, rel=2)
        assert result["upper_bound"] == pytest.approx(0.081661, rel=2)
        assert result["p_values"] == pytest.approx(0.326668, rel=2)

    def test_bootstrap_wrong_variant_column_name(self, click_df_wrong: Callable):
        with pytest.raises(AssertionError):
            compare_bootstrap_delta(
                click_df_wrong, ["control", "experiment"], "click", "impression"
            )

    def test_bootstrap_raise_warning_mismatch_ratio(
        self, click_df_mismatch_ratio: Callable
    ):
        with pytest.warns(UserWarning):
            compare_bootstrap_delta(
                click_df_mismatch_ratio,
                ["control", "experiment"],
                "click",
                "impression",
            )
